#include "MeshLoader.h"
#include "HashCustom.h"

const std::vector<Node>& MeshLoader::getNode() const
{
	return m_node;
}

const std::vector<FiniteElement>& MeshLoader::getFiniteNode() const
{
	return m_fin_el;
}

const std::vector<BoundaryFiniteElement> MeshLoader::getBoundFinNode() const
{
	return m_bound_fin_el;
}

std::set<int> MeshLoader::getNodeBySurface(int surface_id) const
{
	std::set<int> temp;
	auto first_to_find = m_bound_fin_el.begin();
	auto last_to_find = m_bound_fin_el.end();
	while (first_to_find != last_to_find)
	{
		auto found_bound = std::find_if(first_to_find,
											last_to_find,
											[surface_id](const BoundaryFiniteElement& cur_bound) {
													return cur_bound.m_surface_id == surface_id;
											});
		first_to_find = found_bound;
		if (found_bound != last_to_find) {
			for (const auto& node_in_bound : found_bound->m_node_id)
				temp.insert(node_in_bound);
			++first_to_find;
		}
	}
	return temp;
}

std::vector<int> MeshLoader::getElemByID(int node_id1, int node_id2, int node_id3) const
{
	std::vector<int> temp;
	auto first_to_find = m_fin_el.begin();
	auto last_to_find = m_fin_el.end();
	while (first_to_find != last_to_find)
	{
		auto found_elem = std::find_if(first_to_find, last_to_find,
			[node_id1, node_id2, node_id3](const FiniteElement& elem) {
				return std::find(elem.m_node_id.begin(), elem.m_node_id.end(), node_id1) != elem.m_node_id.end()
					&& std::find(elem.m_node_id.begin(), elem.m_node_id.end(), node_id2) != elem.m_node_id.end()
					&& std::find(elem.m_node_id.begin(), elem.m_node_id.end(), node_id3) != elem.m_node_id.end();
			});
		first_to_find = found_elem;
		if (found_elem != last_to_find) {
			temp.push_back(found_elem->m_element_id);
			++first_to_find;
		}
	}
	return temp;
}

std::vector<int>MeshLoader::getElemByEdge(int node_id1, int node_id2) const
{
	std::vector<int> temp;
	auto first_to_find = m_fin_el.begin();
	auto last_to_find = m_fin_el.end();
	while (first_to_find != last_to_find)
	{
		auto found_elem = std::find_if(first_to_find, last_to_find,
			[node_id1, node_id2](const FiniteElement& elem) {
				return std::find(elem.m_node_id.begin(), elem.m_node_id.end(), node_id1) != elem.m_node_id.end()
					&& std::find(elem.m_node_id.begin(), elem.m_node_id.end(), node_id2) != elem.m_node_id.end();
			});
		first_to_find = found_elem;
		if (found_elem != last_to_find) {
			temp.push_back(found_elem->m_element_id);
			++first_to_find;
		}
	}
	return temp;
}

std::vector<int> MeshLoader::getElemByMaterial(int material_id) const
{
	std::vector<int> temp;
	auto first_to_find = m_fin_el.begin();
	auto last_to_find = m_fin_el.end();
	while (first_to_find != last_to_find)
	{
		auto found_elem = std::find_if(first_to_find, last_to_find,
			[material_id](const FiniteElement& elem) {
				return elem.m_element_id == material_id;
			});
		first_to_find = found_elem;
		if (found_elem != last_to_find) {
			temp.push_back(found_elem->m_element_id);
			++first_to_find;
		}
	}
	return temp;
}

std::vector<int> MeshLoader::getBoundBySurface(int surface_id) const
{
	std::vector<int> temp;
	auto first_to_find = m_bound_fin_el.begin();
	auto last_to_find = m_bound_fin_el.end();
	while (first_to_find != last_to_find)
	{
		auto found_elem = std::find_if(first_to_find, last_to_find,
			[surface_id](const BoundaryFiniteElement& elem){
				return elem.m_surface_id == surface_id;
			});
		first_to_find = found_elem;
		if (found_elem != last_to_find) {
			temp.push_back(found_elem->m_element_id);
			++first_to_find;
		}
	}
	return temp;
}

void MeshLoader::insertNode()
{
	std::unordered_set<Edge> test_edge;

	for (auto& elem : m_fin_el) {
		std::vector<int> cur_node_id = elem.m_node_id;

		for (auto first_node = 0; first_node < 4; ++first_node)

			for (auto second_node = first_node + 1; second_node < 4; ++second_node) {
				Edge cur_edge(cur_node_id[first_node], cur_node_id[second_node]);

				if (test_edge.insert(cur_edge).second) {
					Node node = getMidNode(cur_edge);
					cur_edge.update_middle(node.m_node_id);
					m_node.push_back(node);
				}
				else
					elem.m_node_id.push_back(cur_edge.m_middle_node);
			}
	}

	for (auto& elem : m_bound_fin_el) {
		std::vector<int> cur_node_id = elem.m_node_id;

		for (auto first_node = 0; first_node < 3; ++first_node)

			for (auto second_node = first_node + 1; second_node < 3; ++second_node) {
				auto found_edge = test_edge.find({ cur_node_id[first_node], cur_node_id[second_node] });
				if (found_edge == test_edge.end())
					found_edge = test_edge.find({ cur_node_id[second_node], cur_node_id[first_node] });
				elem.m_node_id.push_back(found_edge->m_middle_node);
			}
	}
}

std::vector<std::set<int>> MeshLoader::getAdjacentNode() const
{
	std::vector<std::set<int>> temp(m_node.size() + 1);
	for (const auto& cur_elem : m_fin_el) {
		for (auto cur_node_id : cur_elem.m_node_id)
			for (auto adj_node_id : cur_elem.m_node_id)
				if (cur_node_id != adj_node_id)
					temp[cur_node_id].insert(adj_node_id);
	}
	return temp;
}

std::array<double, 3> MeshLoader::getMidEdge(const Edge& edge)
{
	return
	{ (get<0>(m_node[edge.m_node.first - 1].m_coord) + get<0>(m_node[edge.m_node.second - 1].m_coord)) * (1 / 2)
	, (get<1>(m_node[edge.m_node.first - 1].m_coord) + get<1>(m_node[edge.m_node.second - 1].m_coord)) * (1 / 2)
	, (get<2>(m_node[edge.m_node.first - 1].m_coord) + get<2>(m_node[edge.m_node.second - 1].m_coord)) * (1 / 2) };
}

Node MeshLoader::getMidNode(const Edge& edge)
{
	Node node(m_node.size() + 1, getMidEdge(edge), 0);
	return node;
}

Node& MeshLoader::getNodeByID(int id)
{
	return m_node.at(id);
}

